import gc
import time


class GCRecorder():
    def __init__(self):
        if not hasattr(gc, 'callbacks'):
            self.support = False
            return
        self.support = True
        self.start_time = 0
        self.accumulate_time = 0
        self.gc_count = 0
        self.gen_time = [0, 0, 0]  # Python have 3 generation
        gc.callbacks.append(self.gc_callback)

    def record(self):
        if not self.support:
            return (0, 0, [])
        gc_time = self.accumulate_time
        gc_count = self.gc_count
        gen_time = []
        for t in self.gen_time:
            gen_time.append(t)
        return gc_time, gc_count, gen_time

    def gc_callback(self, phase, info):
        if phase == 'start':
            self.start_time = time.time()
        else:  # phase end
            delta = int((time.time() - self.start_time) * 1000)
            self.accumulate_time += delta
            self.gc_count += info['collected']
            self.gen_time[info['generation']] = delta

    def __del__(self):
        if not self.support:
            return
        if self.gc_callback in gc.callbacks:
            gc.callbacks.remove(self.gc_callback)



import os
import json
import socket
import struct
import logging
import sys
import time
import threading
import traceback
import zlib
from random import random
from jennifer.api import task
from jennifer.recorder import Recorder
from jennifer.api import task
from .transaction import Transaction

record_types = {
    'method': 1,
    'sql': 2,
    'event_detail_msg': 3,
    'service': 4,
    'txcall': 5,
    'browser_info': 6,
    'thread_name': 7,
    'thread_stack': 8,
    'user_id': 9,
    'dbc': 10,
    'stack_trace': 11,
    'count': 12,
}


class Agent(object):
    def __init__(self):
        self.transactions = []
        self.inst_id = -1
        self.recorder = None
        self.logger = logging.getLogger('jennifer')

    def gen_new_txid(self):
        return int(str(int(random() * 100)) + str(self.current_time()))

    def current_time(self):
        return int(time.time() * 1000)

    def current_cputime(self):
        if hasattr(time, 'process_time'):
            return int(time.process_time() * 1000)
        return int(time.clock() * 1000)

    def current_thread_id(self):
        return threading.get_ident()

    def current_thread(self):
        return threading.current_thread()

    def current_transaction(self):
        ret = None
        thread_id = self.current_thread_id()
        for t in self.transactions:
            if t.thread_id == thread_id:
                ret = t
                break
        return ret

    def set_config(self, config):
        self.address = config['address']
        self.recorder = Recorder()
        self.config_pid = os.getpid()
        if self.connect_master():
            task.run_timer(self.agent_loop)

    def connect_master(self):
        self.master = None
        if not os.path.exists(self.address):
            return False
        try:
            self.master = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.master.connect(self.address)
            self.handshake_to_master()
            task.run_task(self.master_loop)
            return True
        except ConnectionRefusedError as e:
            self.master = None
            return False

    def handshake_to_master(self):
        version = b'{"version":1,"pid":' + \
            str(os.getpid()).encode('utf-8') + b'}'
        self.master.send(b'\x08jennifer' + \
            struct.pack('>L', len(version)) + version)

    def send_to_master(self, cmd, params):
        if os.getpid() != self.config_pid:
            self.set_config({
                'address': self.address
            })
        if self.master is None:
            if not self.connect_master():
                return

        try:
            p = json.dumps(params).encode('utf-8')
            pack = struct.pack('>B', len(cmd)) + \
                cmd.encode('utf-8') + \
                struct.pack('>L', len(p)) + \
                p
            self.master.send(pack)
        except (BrokenPipeError, OSError):
            self.master = None
            return

    def agent_loop(self):
        metrics = self.recorder.record_self()
        self.send_to_master('record_metric', metrics)
        current = self.current_time()
        current_cpu = self.current_cputime()
        self.send_to_master('active_service', {
            'active_services': [
                t.to_active_service_dict(current, current_cpu)
                for t in self.transactions
            ],
        })

    def master_loop(self):
        while True:
            if self.master is None:
                break
            try:
                bcmd_len = self.master.recv(1)
                if len(bcmd_len) < 1:
                    break  # failed to connect master
                cmd_len = ord(bcmd_len)
                cmd = self.master.recv(cmd_len)
                param_len, = struct.unpack('>L', self.master.recv(4))
                try:
                    param = json.loads(self.master.recv(param_len))
                except:
                    continue

                if cmd == b'active_detail':
                    txid = param.get('txid')
                    request_id = param.get('request_id')
                    if txid is not None and request_id is not None:
                        data = self.get_active_service_detail(txid)
                        data['request_id'] = request_id
                        self.send_to_master('active_detail', data)
            except (BrokenPipeError, OSError):
                break

    def get_active_service_detail(self, txid):
        ret = None
        cpu_time = self.current_cputime()
        current_time = self.current_time()
        for t in self.transactions:
            if t.txid == txid:
                stack = 'Fail to get a callstack'
                frame = sys._current_frames().get(t.thread_id)
                if frame is not None:
                    stack = ''.join(traceback.format_stack(frame))
                ret = {
                    'txid': t.txid,
                    'service_name': t.path_info,
                    'elapsed': current_time - t.start_time,
                    'method': t.request_method,
                    'http_query': t.query_string,
                    'sql_count': t.sql_count,
                    'sql_time': t.sql_time,
                    'tx_time': t.external_call_time,
                    'tx_count': t.external_call_count,
                    'fetch_count': t.fetch_count,
                    'cpu_time': cpu_time - t.start_system_time,
                    'start_time': t.start_time,
                    'stack': stack,
                }

        return ret

    def start_transaction(
        self,
        environ,
        wmonid,
        wsgi_handler_name=None,
        start_time=None
    ):
        if start_time is None:
            start_time = self.current_time()
        txid = self.gen_new_txid()
        transaction = Transaction(
            agent=self,
            start_time=start_time,
            txid=txid,
            environ=environ,
            wmonid=wmonid,
        )
        # transaction.profiler.set_root_name(wsgi_handler_name)
        self.transactions.append(transaction)
        transaction.start_system_time = self.current_cputime()
        self.send_to_master('start_transaction', {})
        return transaction

    def end_transaction(self, transaction):
        try:
            self.transactions.remove(transaction)
            self.send_to_master('end_transaction', {
                'transaction': transaction.to_dict(),
                'profiler': transaction.profiler.to_dict(),
            })
        except ValueError:
            pass

    def _hash_text(self, text):
        hashkey = zlib.crc32(text.encode('utf-8'))
        if hashkey > 0x7FFFFFFF:
            return (hashkey & 0x7FFFFFFF) * -1
        return hashkey

    def hash_text(self, text, hash_type='service'):
        if text is None or len(text) is 0:
            return 0
        self.send_to_master('record_text', {
            'type': record_types.get(hash_type, 0),
            'text': text,
        })
        return self._hash_text(text)



from .profile_data import ProfileDataType


class File(ProfileDataType):
    TYPE_UNKNOWN = 0
    TYPE_READ = 1
    TYPE_WRITE = 2
    TYPE_RWOPEN = 6

    def __init__(
            self,
            name='',
            mode='',
    ):
        ProfileDataType.__init__(self)
        self.name = name
        is_read = 'r' in mode
        is_write = 'w' in mode

        self.mode = File.TYPE_UNKNOWN
        if is_read and is_write:
            self.mode = File.TYPE_RWOPEN
        elif is_read:
            self.mode = File.TYPE_READ
        elif is_write:
            self.mode = File.TYPE_WRITE

    def get_type(self):
        return ProfileDataType.TYPE_FILE



"""Wsgi Agent for Jennifer APM
"""
import os
import sys
import base64
import struct
from jennifer.agent import jennifer_agent
from email.utils import formatdate
import time
try:
    import Cookie as cookies
except ImportError:
    from http import cookies


wmonid_pack = struct.Struct('>Q')

def wrap_wsgi_start_response(origin, set_wmonid, new_wmonid=None):
    def handler(*args, **kwargs):
        if set_wmonid:
            if len(args) == 2:
                expire = formatdate(
                    timeval=time.time() + 31536000,
                    localtime=False,
                    usegmt=True
                )
                set_cookie = 'WMONID=%s; expires=%s; Max-Age=31536000; path=/' % (
                    base64.b64encode(new_wmonid).decode('ascii'), expire)

                args[1].append(('Set-Cookie', str(set_cookie)))

        return origin(*args, **kwargs)
    return handler

def wrap_wsgi_handler(app, wsgi_handler_name):
    agent = jennifer_agent()

    def handler(*args, **kwargs):
        environ = {}
        wmonid = None
        new_wmonid_val = (os.getpid() << 32) + int(time.time())
        new_wmonid = wmonid_pack.pack(new_wmonid_val)
        if len(args) is 3:
            environ = args[1]  # self, environ, start_response
            modargs = [
                args[0],
                args[1],
            ]
            start_response = args[2]
        elif len(args) is 2:
            environ = args[0]  # environ, start_response
            modargs = [
                args[0],
            ]
            start_response = args[1]
        cookie = cookies.SimpleCookie()
        cookie.load(environ.get('HTTP_COOKIE', ''))
        cookie_wmonid = cookie.get('WMONID')
        if cookie_wmonid is None:
            wmonid = new_wmonid_val
        else:
            try:
                wmonid, = wmonid_pack.unpack(
                    base64.b64decode(cookie_wmonid.value))
            except:  # incorrect wmonid
                cookie_wmonid = None
                wmonid = new_wmonid_val
        modargs.append(
            wrap_wsgi_start_response(
                start_response,
                set_wmonid=(cookie_wmonid == None),
                new_wmonid=new_wmonid,
            )
        )

        transaction = agent.start_transaction(
            environ, wmonid, wsgi_handler_name)
        ret = app(*modargs, **kwargs)
        end_time = agent.current_time()
        transaction.end(end_time)
        return ret
    return handler


def wrap_wsgi_app(app, default_wsgi_handler_name='wsgi_handler'):
    return wrap_wsgi_handler(app, default_wsgi_handler_name)



import sys
import os
from os import path


try:
    jennifer = __import__('jennifer')
except ImportError:
    jennifer_path = path.abspath(path.join(path.dirname(__file__), '..', '..'))
    sys.path.append(jennifer_path)
    jennifer = __import__('jennifer')

if os.environ.get('JENNIFER_MASTER_ADDRESS') is not None:
    config = {
        'address': os.environ['JENNIFER_MASTER_ADDRESS'],
        'log_path': os.environ['JENNIFER_LOG_PATH'],
    }
    jennifer.config.init(config)



from jennifer.agent import jennifer_agent
from jennifer.api.proxy import Proxy


def safe(l, idx, default=None):
    try:
        return l[idx]
    except IndexError:
        return default


class CursorProxy(Proxy):
    def __init__(self, obj, host, port, paramstyle):
        Proxy.__init__(self, obj)
        self.set('host', host)
        self.set('port', port)
        self.set('paramstyle', paramstyle)

    def execute(self, *args, **kwargs):  # down active when sql error occur
        if jennifer_agent().recorder is not None:
            jennifer_agent().recorder.db_recorder.up_active()
        transaction = jennifer_agent().current_transaction()
        operation = safe(args, 0) or kwargs.get('operation')
        parameters = safe(args, 1) or kwargs.get('parameters')
        if transaction is not None and operation is not None:
            if parameters is not None and \
                (self.paramstyle == 'format' or self.paramstyle == 'pyformat'):
                operation = operation % (parameters)
            transaction.profiler.db_execute(self.host, self.port, operation)
        result = None
        try:
            result = self._origin.execute(*args, **kwargs)
        except Exception as e:
            if transaction is not None:
                transaction.profiler.sql_error(e)
            raise e
        if transaction is not None and operation is not None:
            transaction.profiler.end()
        if jennifer_agent().recorder is not None:
            jennifer_agent().recorder.db_recorder.down_active()
        return result

    def record_fetch(self, fetch, size, pass_size=False):
        if jennifer_agent().recorder is not None:
            jennifer_agent().recorder.db_recorder.up_active()
        transaction = jennifer_agent().current_transaction()
        if transaction is None:
            if pass_size:
                ret = fetch(size)
                if jennifer_agent().recorder is not None:
                    jennifer_agent().recorder.db_recorder.down_active()
                return ret
            else:
                ret = fetch()
                if jennifer_agent().recorder is not None:
                    jennifer_agent().recorder.db_recorder.down_active()
                return ret

        if size == 0:  # it mean fetch all
            size = self._result.num_rows()
        transaction.profiler.db_fetch(size)
        if pass_size:
            ret = fetch(size)
        else:
            ret = fetch()
        transaction.profiler.end()
        if jennifer_agent().recorder is not None:
            jennifer_agent().recorder.db_recorder.down_active()
        return ret

    def fetchone(self):
        return self.record_fetch(self._origin.fetchone, 1)

    def fetchmany(self, size=None):
        pass_size = True
        if size is None:
            size = self._origin.arraysize
            pass_size = False
        return self.record_fetch(self._origin.fetchmany, size, pass_size)

    def fetchall(self):
        size = self.rowcount
        return self.record_fetch(self._origin.fetchall, size)


class ConnectionProxy(Proxy):
    def __init__(self, obj, host, port, paramstyle):
        Proxy.__init__(self, obj)
        self.set('host', host)
        self.set('port', port)
        self.set('paramstyle', paramstyle)

    def cursor(self, *args, **kwargs):
        cursor = self._origin.cursor(*args, **kwargs)
        return CursorProxy(cursor, self.host, self.port, self.paramstyle)

    def close(self, *args, **kwargs):
        if jennifer_agent().recorder is not None:
            jennifer_agent().recorder.db_recorder.down_total()
        return self._origin.close(*args, **kwargs)


def register_database(module, connection_info):
    agent = jennifer_agent()
    def wrap_connect(connect):
        def handler(*args, **kwargs):
            host, port, database = connection_info(*args, **kwargs)
            transaction = agent.current_transaction()
            if agent.recorder is not None:
                agent.recorder.db_recorder.up_total()
            if transaction is not None:
                transaction.profiler.db_open(host, port, database)
            try:
                connection = ConnectionProxy(
                    connect(*args, **kwargs), host, port, module.paramstyle)
            except Exception as e:
                transaction = agent.current_transaction()
                if transaction is not None:
                    transaction.profiler.db_connection_error(e)
                raise e
            if transaction is not None:
                transaction.profiler.end()
            return connection
        return handler
    module.connect = wrap_connect(module.connect)



import os
import sys
import time
import platform
import tempfile
import subprocess
from configparser import ConfigParser


def print_help():
    print("""
Usage: jennfier run <startup_command ...>

startup_command: your wsgi service startup command and options
""")


def run_master(bin, config_path, log_path, sock_file):
    arch = {
        'x86_64': 'amd64',
        'x86': '386',
    }[platform.machine()]

    path = os.path.join(bin, sys.platform, arch, 'jennifer_agent')
    log_stream  = open(log_path, 'w+')
    process = subprocess.Popen(
        [
            path,
            config_path,
            sock_file,
        ],
        stdout=log_stream,
        stderr=log_stream,
    )

def run(args):
    if len(args) < 2:
        print_help()
        return
    startup_path = args[1]
    config_path = os.environ.get('JENNIFER_CONFIG_FILE')
    if config_path is None:
        raise EnvironmentError("JENNIFER_CONFIG_FILE is not set")
    config_path = os.path.join(os.getcwd(), config_path)
    if not os.path.exists(config_path):
        raise FileNotFoundError("JENNIFER_CONFIG_FILE not exist")

    root_dir = os.path.dirname(__import__('jennifer').__file__)
    bootstrap = os.path.join(root_dir, 'bootstrap')

    python_path = bootstrap
    if 'PYTHONPATH' in os.environ:
        path = os.environ['PYTHONPATH'].split(os.path.pathsep)
        if bootstrap not in path:
            python_path = os.path.pathsep.join([bootstrap] + path)
    os.environ['PYTHONPATH'] = python_path

    if not os.path.dirname(startup_path):
        for path in os.environ.get('PATH', '').split(os.path.pathsep):
            path = os.path.join(path, startup_path)
            if os.path.exists(path) and os.access(path, os.X_OK):
                startup_path = path
                break

    if not os.path.exists(startup_path):
        raise FileNotFoundError('{0} not found'.format(startup_path))

    sock = os.path.join(tempfile.gettempdir(), 'jennifer-%d.sock' % time.time())
    os.environ['JENNIFER_MASTER_ADDRESS'] = sock
    config = ConfigParser()
    config.read(config_path)
    log_path = config['JENNIFER'].get(
        'log_path', '/tmp/jennifer-python-agent.log')
    os.environ['JENNIFER_LOG_PATH'] = log_path

    # Start!
    pid = os.fork()
    if pid > 0:  # parent process
        os.waitpid(pid, 0)
        os.execl(startup_path, *args[1:])
    else:  # child process
        run_master(os.path.join(root_dir, 'bin'), config_path, log_path, sock)



import time


class ProfileDataType:
    TYPE_EMPTY = 0
    TYPE_METHOD = 0x10
    TYPE_TXCALL = 0x20
    TYPE_SQL_EXEC = 0x30
    TYPE_SQL_FETCH = 0x40
    TYPE_MESSAGE = 0x50
    TYPE_DBMESSAGE = 0x60
    TYPE_FILE = 8 << 4
    TYPE_SOCKET = 9 << 4
    TYPE_EXCEPTION = 10 << 4
    TYPE_METHOD_PARAM = 13 << 4
    TYPE_METHOD_RETURN = 14 << 4
    TYPE_ERROR = 1

    def __init__(self):
        self.parent_index = 0
        self.index = 0
        self.start_time = 0
        self.start_cpu = 0
        self.elapsed_time = 0
        self.elapsed_cpu = 0
        self.parent = None

    def get_type(self):
        return ProfileDataType.TYPE_EMPTY

    def content_bytes(self):
        return b''

    def index_to_bytes(self):
        return raw.pack_with(
            raw.from_number(self.index),
            raw.from_number(self.parent_index),
        )

    def to_dict(self):
        ret = self.__dict__
        ret['type'] = self.get_type()
        ret.pop('parent', None)  # clear parent
        return ret


class ProfileData:
    def __init__(self, txid, service_hash, children):
        self.txid = txid
        self.service_hash = service_hash
        self.children = children

    def to_dict(self):
        return {
            'txid': self.txid,
            'service_hash': self.service_hash,
            'profiles': [d.to_dict() for d in self.children],
        }



import sys
import logging
from jennifer.api.task import force_shutdown
from jennifer.agent import jennifer_agent
from jennifer.hooks import hooking
from jennifer.api.config import ConfigNotExistsError


def setup_log(config):
    logger = logging.getLogger('jennifer')
    logger.setLevel(logging.INFO)
    logger.propagate = False
    handler = logging.FileHandler(config['log_path'])
    formatter = logging.Formatter(
        '%(asctime)s [JENNIFER Python] %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

def init(config):
    agent = jennifer_agent()
    agent.set_config(config)
    hooking()



import os
import time
import traceback
from threading import Thread, Timer


def run_task(target, args=()):
    t = Thread(target=target, args=args)
    t.daemon = True
    t.start()
    return t


def run_task_safe(target, args, err_handler):
    def _target(*args, **kwargs):
        try:
            target(*args, **kwargs)
        except Exception as e:
            err_handler(e, traceback.format_exc())
    return run_task(_target, args)


def run_timer(target, interval=1):
    def handler():
        while True:
            target()
            time.sleep(interval)
    t = Thread(target=handler)
    t.daemon = True
    t.start()
    return t


def run_timer_safe(target, err_handler, interval=1):
    def handler():
        while True:
            try:
                target()
            except Exception as e:
                err_handler(e, traceback.format_exc())
                return
            time.sleep(interval)
    t = Thread(target=handler)
    t.daemon = True
    t.start()
    return t

def force_shutdown(status_code=-1):
    os._exit(status_code)



from .profile_data import ProfileDataType


class Method(ProfileDataType):
    def __init__(
            self,
            name_hash=0,
            error_hash=0,
    ):
        ProfileDataType.__init__(self)
        self.name_hash = name_hash
        self.error_hash = error_hash

    def get_type(self):
        return ProfileDataType.TYPE_METHOD



from jennifer.api.format import format_function
from jennifer.agent import jennifer_agent
from jennifer.wrap.wsgi import wrap_wsgi_app

__hooking_module__ = 'flask'

def wrap_error_handler(origin):
    try:
        from werkzeug.exceptions import InternalServerError, NotFound
    except ImportError:
        InternalServerError = None
        NotFound = None

    def handler(*args, **kwargs):
        e = None
        if len(args) >= 2:
            e = args[1]
        else:
            e = kwargs.get('e')
        if e is not None:
            profiler = jennifer_agent().current_transaction().profiler
            if type(e) == InternalServerError:
                profiler.service_error(e)
            elif type(e) == NotFound:
                profiler.not_found(e)
            else:
                profiler.exception(e)
        return origin(*args, **kwargs)
    return handler

def wrap_dispatch_requet(origin, flask):
    def handler(self):
        req = flask.ctx._request_ctx_stack.top.request
        if req is not None and req.url_rule is not None:
            handler = self.view_functions.get(req.url_rule.endpoint, None)
            handler_name = format_function(handler)
            agent = jennifer_agent()
            transaction = agent.current_transaction()
            if transaction is not None and handler_name != '':
                transaction.profiler.set_root_name(handler_name)
        return origin(self)
    return handler

def hook(flask):
    flask.Flask._find_error_handler = wrap_error_handler(
        flask.Flask._find_error_handler)
    flask.Flask.wsgi_app = wrap_wsgi_app(
        flask.Flask.wsgi_app,
        'flask.Flask.wsgi_app')
    flask.Flask.dispatch_request = wrap_dispatch_requet(
        flask.Flask.dispatch_request, flask)



from jennifer.wrap import db_api
from jennifer.agent import jennifer_agent
__hooking_module__ = 'sqlite3'


def connection_info(database, *args, **kwargs):
    return 'localhost', 0, database


def hook(sqlite3):
    db_api.register_database(sqlite3.dbapi2, connection_info)



from .method import Method
from .profile_data import ProfileData
from .socket import Socket
from .file import File
from .external_call import ExternalCall
from .message import Message
from .root_method import Root
from .error import Error
from . import sql



from ..profile_data import ProfileDataType


class Message(ProfileDataType):
    TYPE_UNKNOWN = 0
    TYPE_OPEN = 1
    TYPE_CLOSE = 2
    TYPE_COMMIT = 3
    TYPE_ROLLBACK = 4
    TYPE_AUTO_COMMIT_FAILED = 5

    def __init__(
            self,
            txid=0,
            message='',
            message_type=TYPE_UNKNOWN,
    ):
        ProfileDataType.__init__(self)
        self.message = message
        self.message_type = message_type
        self.txid = txid

    def get_type(self):
        return ProfileDataType.TYPE_DBMESSAGE



class DBConnectionRecorder(object):
    def __init__(self):
        self.total = 0
        self.active = 0

    def up_total(self):
        self.total += 1

    def down_total(self):
        self.total -= 1

    def up_active(self):
        self.active += 1

    def down_active(self):
        self.active -= 1

    def record(self):
        return (
            self.total,
            self.active,
        )



import os
import jennifer
from distutils.command.sdist import sdist as _sdist
from setuptools import setup, find_packages


class sdist(_sdist):
    def run(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        support_platforms = [
            # ['darwin', '386'],
            ['darwin', 'amd64'],
            # ['linux', '386'],
            ['linux', 'amd64'],
        ]
        os.system('rm -rf %s/jennifer/bin' % dir_path)
        for platform in support_platforms:
            os.system(
                ("env GOOS={0} GOARCH={1} go build -o " + \
                "{2}/jennifer/bin/{0}/{1}/jennifer_agent jennifer").format(
                    platform[0],
                    platform[1],
                    dir_path,
                )
            )
        _sdist.run(self)


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name="jennifer",
    description="JENNIFER, JenniferSoft APM, python agent.",
    long_description=readme(),
    cmdclass={'sdist': sdist},
    version=jennifer.__version__,
    author=jennifer.__author__,
    author_email="python@jennifersoft.com",
    url="http://jennifersoft.com",
    license="Proprietary",
    packages=find_packages(exclude=[]),
    package_data={
        'jennifer': ['bin/*/*/jennifer_agent'],
    },
    entry_points = {
        'console_scripts': [
            'jennifer-admin = jennifer.admin:main',
        ],
    },
    classifiers=[
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'License :: Other/Proprietary License',
    ],
)
